'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _babelTraverse = require('babel-traverse');

var _babelTraverse2 = _interopRequireDefault(_babelTraverse);

var _builtin = require('./builtin');

var builtin = _interopRequireWildcard(_builtin);

var _visitors = require('./visitors');

var _utils = require('./utils');

var _symbols = require('./symbols');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Macro = function () {
  function Macro(_ref3) {
    var name = _ref3.name,
        macroBody = _ref3.macroBody,
        scope = _ref3.scope,
        location = _ref3.location;

    _classCallCheck(this, Macro);

    location = location || '';
    (0, _babelTraverse2.default)(macroBody, _babelTraverse2.default.visitors.merge([_visitors.collectMacros, {
      ThisExpression: function ThisExpression() {
        throw new Error("Can not use `this` in macro" + location);
      },
      Identifier: function Identifier(_ref4) {
        var node = _ref4.node;

        if ("arguments" === node.name) {
          throw new Error("Can not use `arguments` in macro" + location);
        }
      }
    }]), scope);
    this.name = name;
    this.macroBody = macroBody;
    this.scope = scope;
  }

  _createClass(Macro, [{
    key: 'run',
    value: function run(path, scope) {
      if (!this.macro) {
        this.macro = Macro.compileMacro(this.name, this.macroBody, this.scope);
      }
      return this.macro(path, scope);
    }
  }], [{
    key: 'register',
    value: function register(name, macroPath, scope, location) {
      var macroBody = macroPath.node,
          subScope = macroPath.scope;
      location = location || '';
      scope[_symbols.$registeredMacros] = scope[_symbols.$registeredMacros] || {};
      scope[_symbols.$registeredMacros][name] = new Macro({ name: name, macroBody: macroBody, scope: subScope, location: location });
    }
  }, {
    key: 'compileMacro',
    value: function compileMacro(name, node, scope) {
      if (!(typeof name === 'string')) {
        throw new TypeError('Value of argument "name" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(name));
      }

      if (!(node instanceof Object)) {
        throw new TypeError('Value of argument "node" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(node));
      }

      if (!(scope instanceof Object)) {
        throw new TypeError('Value of argument "scope" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(scope));
      }

      (0, _babelTraverse2.default)(node, _visitors.processMacros, scope);

      var _collectReferences = (0, _utils.collectReferences)(node, scope),
          references = _collectReferences.references,
          paramReferenceCounts = _collectReferences.paramReferenceCounts;

      return function (path, scope) {
        var cloned = t.cloneDeep(node);
        var uid = scope.generateUidIdentifier((0, _utils.camelCase)(name));
        var labelUid = scope.generateUidIdentifier('_' + name.toUpperCase());
        var argsMacros = {};

        var _cloned$params$reduce = cloned.params.reduce(function (_ref5, id, index) {
          var _ref6 = _slicedToArray(_ref5, 2),
              params = _ref6[0],
              seen = _ref6[1];

          var argument = path.get('arguments.' + index);
          if (argument && argument.isFunction()) {
            argsMacros[id.name] = new Macro({ name: id.name, macroBody: argument.node, scope: argument.scope });
          } else {
            params[id.name] = {
              id: id,
              replacement: path.node.arguments[index] || t.identifier('undefined'),
              reference: null
            };
            seen[id.name] = false;
          }
          return [params, seen];
        }, [{}, {}]),
            _cloned$params$reduce2 = _slicedToArray(_cloned$params$reduce, 2),
            params = _cloned$params$reduce2[0],
            seen = _cloned$params$reduce2[1];

        (0, _babelTraverse2.default)(cloned, _visitors.processMacros, scope, argsMacros, path.parentPath);

        var hasMultipleReturn = (0, _utils.checkMultipleReturn)(cloned, scope);
        (0, _babelTraverse2.default)(cloned, {
          Identifier: function Identifier(subPath) {
            var child = subPath.node,
                parent = subPath.parent;

            if (parent.type !== "MemberExpression" || parent.object === child || parent.computed && parent.property === child) {
              if (params[child.name]) {
                var param = params[child.name];
                if (param.replacement.type === 'Identifier' || param.replacement.type === 'Literal' || paramReferenceCounts[child.name] === 1 && param.replacement.type === 'MemberExpression') {
                  subPath.replaceWith(param.replacement);
                  seen[child.name] = param.replacement;
                } else {
                  if (!seen[child.name]) {
                    seen[child.name] = scope.generateUidIdentifier(child.name);

                    (0, _utils.getParentBlock)(path).insertBefore([t.variableDeclaration('const', [t.variableDeclarator(seen[child.name], param.replacement)])]);
                  }
                  subPath.replaceWith(seen[child.name]);
                }
              } else if (references[child.name]) {
                if (!seen[child.name]) {
                  seen[child.name] = scope.generateUidIdentifier(child.name);
                }
                subPath.replaceWith(seen[child.name]);
              }
            }
          },
          ReturnStatement: function ReturnStatement(subPath) {
            var child = subPath.node;

            var isLast = child === cloned.body.body[cloned.body.body.length - 1];
            subPath.replaceWith(t.expressionStatement(t.assignmentExpression('=', uid, child.argument || t.identifier('undefined'))));
            if (hasMultipleReturn && !isLast) {
              subPath.insertAfter(t.breakStatement(labelUid));
            }
          },
          FunctionDeclaration: function FunctionDeclaration() {
            // @todo need correct rename. now renames only usages, but not function
            throw new Error('FunctionDeclaration in macros are not supported temporarily');
          },
          Function: function Function(subPath) {
            subPath.skip();
          }
        }, scope);

        if (t.isStatement(cloned.body)) {
          var parentBlock = (0, _utils.getParentBlock)(path);
          parentBlock.insertBefore([t.variableDeclaration('let', [t.variableDeclarator(uid)])]);

          if (hasMultipleReturn) {
            parentBlock.insertBefore(t.labeledStatement(labelUid, cloned.body));
          } else {
            parentBlock.insertBefore(cloned.body.body);
          }
          path.replaceWith(uid);
        } else {
          path.replaceWith(cloned.body);
        }
      };
    }
  }, {
    key: 'runMacro',
    value: function runMacro(path, macro, scope) {
      if (typeof macro === 'function') {
        macro(path, scope);
      } else {
        macro.run(path, scope);
      }
    }
  }, {
    key: 'getMacro',
    value: function getMacro(node, scope, isBuiltinMacro) {
      function _ref2(_id2) {
        if (!(_id2 instanceof Macro || typeof _id2 === 'function' || _id2 == null)) {
          throw new TypeError('Function return value violates contract.\n\nExpected:\nMacro | Function | void\n\nGot:\n' + _inspect(_id2));
        }

        return _id2;
      }

      if (!(node instanceof Object)) {
        throw new TypeError('Value of argument "node" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(node));
      }

      if (!(scope instanceof Object)) {
        throw new TypeError('Value of argument "scope" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(scope));
      }

      if (!(typeof isBuiltinMacro === 'boolean' || isBuiltinMacro instanceof Object)) {
        throw new TypeError('Value of argument "isBuiltinMacro" violates contract.\n\nExpected:\nboolean | Object\n\nGot:\n' + _inspect(isBuiltinMacro));
      }

      if (node.type === 'CallExpression') {
        return _ref2(Macro.getMacro(node.callee, scope, isBuiltinMacro));
      } else if (t.isIdentifier(node)) {
        if (isBuiltinMacro) {
          if ('object' === (typeof isBuiltinMacro === 'undefined' ? 'undefined' : _typeof(isBuiltinMacro))) {
            if (isBuiltinMacro[node.name]) {
              return _ref2(isBuiltinMacro[node.name]);
            }
          } else if (builtin[node.name]) {
            return _ref2(builtin[node.name]);
          }
          return;
        }
        while (scope) {
          if (scope[_symbols.$registeredMacros] && scope[_symbols.$registeredMacros][node.name]) {
            return _ref2(scope[_symbols.$registeredMacros][node.name]);
          }
          scope = scope.parent;
        }
      } else if (t.isMemberExpression(node) && !node.computed && t.isIdentifier(node.property)) {
        return _ref2(Macro.getMacro(node.property, scope, isBuiltinMacro));
      }
    }
  }]);

  return Macro;
}();

exports.default = Macro;
;

function _inspect(input, depth) {
  var maxDepth = 4;
  var maxKeys = 15;

  if (depth === undefined) {
    depth = 0;
  }

  depth += 1;

  if (input === null) {
    return 'null';
  } else if (input === undefined) {
    return 'void';
  } else if (typeof input === 'string' || typeof input === 'number' || typeof input === 'boolean') {
    return typeof input === 'undefined' ? 'undefined' : _typeof(input);
  } else if (Array.isArray(input)) {
    if (input.length > 0) {
      if (depth > maxDepth) return '[...]';

      var first = _inspect(input[0], depth);

      if (input.every(function (item) {
        return _inspect(item, depth) === first;
      })) {
        return first.trim() + '[]';
      } else {
        return '[' + input.slice(0, maxKeys).map(function (item) {
          return _inspect(item, depth);
        }).join(', ') + (input.length >= maxKeys ? ', ...' : '') + ']';
      }
    } else {
      return 'Array';
    }
  } else {
    var keys = Object.keys(input);

    if (!keys.length) {
      if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
        return input.constructor.name;
      } else {
        return 'Object';
      }
    }

    if (depth > maxDepth) return '{...}';
    var indent = '  '.repeat(depth - 1);
    var entries = keys.slice(0, maxKeys).map(function (key) {
      return (/^([A-Z_$][A-Z0-9_$]*)$/i.test(key) ? key : JSON.stringify(key)) + ': ' + _inspect(input[key], depth) + ';';
    }).join('\n  ' + indent);

    if (keys.length >= maxKeys) {
      entries += '\n  ' + indent + '...';
    }

    if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
      return input.constructor.name + ' {\n  ' + indent + entries + '\n' + indent + '}';
    } else {
      return '{\n  ' + indent + entries + '\n' + indent + '}';
    }
  }
}