'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getParentBlock = getParentBlock;
exports.getParentScope = getParentScope;
exports.camelCase = camelCase;
exports.checkMultipleReturn = checkMultipleReturn;
exports.collectReferences = collectReferences;
exports.warning = warning;
exports.getLocationMessage = getLocationMessage;

var _babelTraverse = require('babel-traverse');

var _babelTraverse2 = _interopRequireDefault(_babelTraverse);

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function getParentBlock(path) {
  "use strict";

  while (path.parentPath.type !== 'Program' && path.parentPath && !(path.parentPath.isLabeledStatement() || path.parentPath.isBlockStatement())) {
    path = path.parentPath;
  }
  return path;
};

function getParentScope(path) {
  "use strict";

  while (path.parentPath.type !== 'Program' && path.parentPath && !path.parentPath.isStatementOrBlock()) {
    path = path.parentPath;
  }
  return path.scope;
};

function camelCase(input) {
  "use strict";

  return input.toLowerCase().replace(/_(.)/g, function (match, char) {
    return char.toUpperCase();
  });
};

function checkMultipleReturn(node, scope) {
  "use strict";

  var hasMultipleReturn = false;
  (0, _babelTraverse2.default)(node, {
    ReturnStatement: function ReturnStatement(path, state) {
      if (++state.count > 1) {
        hasMultipleReturn = true;
        path.stop();
      }
    }
  }, scope, { count: 0 });
  return hasMultipleReturn;
};

function collectReferences(node, scope) {
  "use strict";

  var paramNames = node.params.map(function (param) {
    return param.name;
  });
  var paramReferenceCounts = {};
  var references = Object.create(null);
  (0, _babelTraverse2.default)(node, {
    enter: function enter(subPath) {
      var child = subPath.node,
          parent = subPath.parent;

      if (subPath.isVariableDeclarator() || subPath.isFunctionDeclaration()) {
        references[child.id.name] = true;
      } else if (subPath.isIdentifier() && (!t.isFunction(parent) || parent.type === "ArrowFunctionExpression" && parent.body === child) && (!t.isMemberExpression(parent) || parent.object === child) && ~paramNames.indexOf(child.name)) {
        paramReferenceCounts[child.name] = paramReferenceCounts[child.name] || 0;
        paramReferenceCounts[child.name]++;
      }
    }
  }, scope);
  return { paramReferenceCounts: paramReferenceCounts, references: references };
};

function warning(msg) {
  "use strict";

  console.error(msg);
};

function getLocationMessage(node) {
  "use strict";

  return node.loc ? ', at Line: ' + node.loc.start.line + ' Column: ' + (node.loc.start.column + 1) : '';
};